from apm import *
import numpy as np
import random

def sim_init(s,a):
   apm(s,a,'clear all')
   apm_load(s,a,'process.apm')
   csv_load(s,a,'process.csv')
   apm_option(s,a,'nlc.imode',4)
   apm_option(s,a,'nlc.nodes',3)
   apm_info(s,a,'SV','y')
   apm_info(s,a,'FV','u')
   apm_option(s,a,'u.fstatus',1)
   msg = 'Successful simulator initialization'
   return msg
   
def mhe_init(s,a):
   apm(s,a,'clear all')
   apm_load(s,a,'model.apm')
   csv_load(s,a,'data.csv')
   apm_option(s,a,'nlc.imode',5)
   apm_option(s,a,'nlc.nodes',3)
   apm_option(s,a,'nlc.web_plot_freq',1)
   apm_info(s,a,'FV','K')
   apm_info(s,a,'FV','tau')
   apm_info(s,a,'MV','u')
   apm_info(s,a,'CV','y')
   # status, whether the optimizer can use it
   apm_option(s,a,'K.status',1)
   apm_option(s,a,'tau.status',1)
   apm_option(s,a,'u.status',0)
   apm_option(s,a,'K.dcost',0.01)
   apm_option(s,a,'tau.dcost',0.01)
   apm_option(s,a,'K.dmax',10)
   apm_option(s,a,'tau.dmax',10)
   # feedback status
   apm_option(s,a,'K.fstatus',0)
   apm_option(s,a,'tau.fstatus',0)
   apm_option(s,a,'u.fstatus',1)
   apm_option(s,a,'y.fstatus',1)
   msg = 'Successful estimator initialization'   
   return msg

def mpc_init(s,a):
   apm(s,a,'clear all')
   apm_load(s,a,'model.apm')
   csv_load(s,a,'data.csv')
   apm_option(s,a,'nlc.imode',6)
   apm_option(s,a,'nlc.nodes',3)
   apm_option(s,a,'nlc.web_plot_freq',1)
   apm_info(s,a,'FV','K')
   apm_info(s,a,'FV','tau')
   apm_info(s,a,'MV','u')
   apm_info(s,a,'CV','y')
   # status, whether the optimizer can use it
   apm_option(s,a,'K.status',0)
   apm_option(s,a,'tau.status',0)
   apm_option(s,a,'u.status',1)
   apm_option(s,a,'y.status',1)
   # feedback status
   apm_option(s,a,'K.fstatus',1)
   apm_option(s,a,'tau.fstatus',1)
   apm_option(s,a,'u.fstatus',0)
   apm_option(s,a,'y.fstatus',1)
   # constraints
   apm_option(s,a,'u.upper',100)
   apm_option(s,a,'u.lower',0)
   # reference trajectory tuning
   apm_option(s,a,'nlc.traj_init',2)
   apm_option(s,a,'nlc.traj_open',0.5)
   apm_option(s,a,'y.tau',10)
   msg = 'Successful controller initialization'
   return msg
   
def sim(s,a,u):
   apm_meas(s,a,'u',u)
   apm(s,a,'solve')
   y = apm_tag(s,a,'y.model')
   return y

def mhe(s,a,inputs):
   y_meas = inputs[0]
   u_meas = inputs[1]
   apm_meas(s,a,'y',y_meas)
   apm_meas(s,a,'u',u_meas)
   apm(s,a,'solve')
   z = apm_sol(s,a)
   K = z['k'][-1]
   tau = z['tau'][-1]
   y_pred = z['y'][-1]
   return [K,tau,y_pred]
   
def mpc(s,a,inputs):
   sp = inputs[0]
   y_meas = inputs[1]
   k_pred = inputs[2]
   tau_pred = inputs[3]
   y_pred = inputs[4]
   apm_meas(s,a,'y',y_meas)
   apm_meas(s,a,'k',k_pred)
   apm_meas(s,a,'tau',tau_pred)
   sphi = sp + 0.1
   splo = sp - 0.1
   apm_option(s,a,'y.sphi',sphi)
   apm_option(s,a,'y.splo',splo)
   apm_option(s,a,'y.sp',sp)
   apm(s,a,'solve')
   u = apm_tag(s,a,'u.newval')
   return u
   

# Number of cycles to run
cycles = 50

# Pulse in the process setpoint
u = 0
sp = 5

# Server
s = 'http://byu.apmonitor.com'
# Application names
a1 = 'sim' + str(int(random.random()*10000))
a2 = 'mhe' + str(int(random.random()*10000))
a3 = 'mpc' + str(int(random.random()*10000))
# Initialize applications
print(sim_init(s,a1))
print(mhe_init(s,a2))
print(mpc_init(s,a3))

# Run cycles
for i in range(cycles):
    # Display cycle
    print('Cycle: ' + str(i+1))
    
    ## Change setpoint
    if i==19:
        sp = 10
    elif i==39:
        sp = 0
    
    ## Process simulator
    measurement_noise = 0.1*(random.random()-0.5)
    y_meas = sim(s,a1,u) + measurement_noise
    
    ## Moving Horizon Estimator (MHE)
    [k_pred,tau_pred,y_pred] = mhe(s,a2,[y_meas,u])
    
    ## Model Predictive Control (MPC)
    u = mpc(s,a3,[sp,y_meas,k_pred,tau_pred,y_pred])
	
    if (i==0):
       # Web viewers to see solution progression
       apm_web(s,a2) # Estimator
       apm_web(s,a3) # Controller
